#!/usr/bin/perl
#============================================================= -*-perl-*-
#
# jsecret.pl:
#
# DESCRIPTION
#   Print out secret.pdb and secret.dat databases
#   
# AUTHOR
#   Jeff Kosowsky
#
#========================================================================
#
# COPYRIGHT 2008, 2009
# Version 0.3, released Nov 2009
# Version 0.4
#  az Thu Sep  8 15:03:06 2011
#  extended by Alexander Zangerl (az@snafu.priv.at) to handle passwords
#  between 17 and 24 characters long
#
#========================================================================
#
# NOTE: This version works on both .dat and .pdb files since it matches 
# for raw records.
# In particular, it does not need the the Palm::PDB or Palm::Raw libraries
# The following header structure is assumed in order to identify each record
#	Header length is 38 bytes, consisting of:
#     4 byte start string: \xFF\@D[\cC\cb]
#     21 byte category (up to 20 characters with NULL terminator (and padding))
#     8 byte signature (IDEA encryption of first 8 bytes of garbled password
#     1 byte TAN mode flag (off/on)
#     4 byte length (32-bit long, big-endian) for size of following data
#
#========================================================================

use strict;
use warnings;

use Crypt::IDEA;
use Term::ReadKey;
#use Data::Dumper; #Just for debugging...

use constant DATA_START  => 38;

my $list;
if ($ARGV[0] =~ /-l/ ) {
	$list=1;
	shift @ARGV;
}
usage() unless $#ARGV >= 0 && $#ARGV <=1;

my $passwd;
ReadMode('noecho');
while(1) {
	printf STDERR "Please enter password: ";
	$passwd = ReadLine(0);
	chop($passwd);
	printf STDERR "\n";
	last if $passwd ne '' || $list;
}
ReadMode('restore');

my $cipher;
my $signature="";
if($passwd) {
	$passwd = uc($passwd) . "\0" x (8 - length($passwd)); #Upcase & NULL pad
	$cipher = new IDEA(garble($passwd)); #Note cipher uses garbled signature
	$signature = $cipher->encrypt(substr($passwd,0,8)); #But signature doesn't
}

my $recordsref = read_records($ARGV[0], $signature);
die "No records match password...\n" unless $#{$recordsref} >= 0;

my $regex=(defined $ARGV[1] ? qr/$ARGV[1]/ : undef);
my $regmatch=0;
foreach my $record (@$recordsref) {
	my $length = get_record_info($record, \ (my $category), \ (my $tanflag));;
	next if defined($regex) && $category !~ /$regex/; #No category match
	$regmatch++;
	if ($list) {print "$category [$length, $tanflag]\n"; next};
	print "**********$category [$length, $tanflag]*********\n";
#	print "*******$category*******\n";  #Substituting this gives same format as the originals scat3 program
	print decrypt($record, $length, $cipher) . "\n";
}
warn "No record categories match regex '$ARGV[1]'\n" 
	if defined($regex) && !$regmatch;
exit;

###########################################################################
sub read_records {
	my ($file, $signature) = @_;
	my $data;

	if (open(FD, "<$file")) {
		local $/;
		$data = <FD>;
		close FD;
	}
	else {die "Error: Can't open $file...\n"; }

	my @records = split(/(?=\xFF\@D[\cC\cb].{20}\0\Q$signature\E)/so, $data);
	#Note including the signature means that we only match on
	#decodable records. Undecodable records will be tacked on to the
	#end of decodable ones, but that won't hurt anything since length
	#is decoded from header
	#Note \Q..\E construct stops interpreting special chars within $signature

	shift @records; # Drop data before first record
	return \@records;
}

#Retrieve category, tanflag and length of record
sub get_record_info {
	my($rawdata, $categoryref, $tanflagref) = @_;
	$rawdata =~ /^\xFF\@D[\cC\cb]([^\0]*)\0/; 
	$$categoryref = $1;
	$$tanflagref = unpack 'c', substr($rawdata, 33, 1); # TAN mode off/on???
	return unpack("N", substr($rawdata, 34, 4));
}

#Depcrypt text of record
sub decrypt {
	my ($rawdata, $length, $cipher) = @_;
	my $rawtext=substr($rawdata,DATA_START); #Rest is the encrypted data
	my $text = "";
	for(my $i=0; $i< $length; $i+=8) {  # 8-byte blocksize
		$text .= $cipher->decrypt(substr($rawtext,$i,8));
	}
	return substr($text,0,$length);
}

# Garble password as required by Secret!
# idea algorithm needs key of precisely 16 bytes, and garble expands/condenses passwords 
# between 1 and 24 bytes to this 16 byte key
#
# priming: all key slots start with zero, then password is filled in: even slots hold pass[0,1,2...]; 
#  odd slots have pass[i-3] and slot 1 stays zero. stop when out of password chars (ie. shorter than 9 chars).
# password between 9 and 16 chars: primed like above, then even slots xor'd with pass[8], pass[9]...; 
#  odd slots xor'd with slot[i-3] and slot 1 is still zero.
# password between 17 and 24 chars: primed like before, then slot 0 is xor'd with low_nibble(pass[16])
#  other even slots xor'd with (high_nibble(slot[i-1]) or low_nibble(pass[17,18...]));
#  odd slots are xor'd with high_nibble(pass[16,17,18...])
sub garble { 

	my @chars = unpack('C*',$passwd);
	my @key = (0) x 16;

	my $pwlen=@chars;
	for my $i (0..7)
	{
	    if ($i+16<$pwlen) 	# 16+ chars long
	    {
		$key[2*$i+1]= ($i==0? 0: $chars[7+$i]) ^ (0xff & ($chars[16+$i]<<4));
		$key[2*$i] = $chars[$i] ^ $chars[8+$i] ^ ($i==0? ($chars[16+$i]>>4) : (($chars[16+$i]>>4)|(($key[2*$i-1]<<4)&0xff)));
	    }
	    elsif ($i+8<$pwlen) # 8+ chars long
	    {
		$key[2*$i]=$chars[$i] ^ $chars[8+$i];
		$key[2*$i+1]=$chars[7+$i] unless ($i==0); # key[1] is fixed at 0
	    }
	    else # 1..7 chars long
	    {
		$key[2*$i]=$chars[$i];
		$key[2*$i+1]=$chars[$i-1] unless ($i==0); # key[1] is fixed at 0
	    }
	}	
	return  pack('C16',@key);
}

sub usage {
	print <<EOF;

  Usage: $0 [-l] <secret pdb or dat file> [regexp]

  where:
    -l means list categories only (if no password given then list all)
    [regexp] means only print categories matching 'regexp'

  Note: The category name is followed by a bracketed pair consisting of the
  record length (in chars) and the TAN mode flag (in decimal).
EOF
exit 0;
}
